/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your option) any
 * later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */
package org.processmining.xeslite.external;

import java.util.Collection;
import java.util.Map;
import java.util.NavigableMap;
import java.util.Set;
import java.util.concurrent.ConcurrentNavigableMap;

import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.impl.XAttributeMapImpl;

import com.google.common.base.Function;
import com.google.common.collect.Collections2;

/**
 * Implements the {@link XAttributeMap} interface using a
 * {@link ConcurrentNavigableMap} from MapDB as backend. This implementation
 * tries to forward most operations directly to the backing map to avoid
 * unnecessary copies. An exception is the method {@link #entrySet()} that
 * returns an immutable copy of the attributes, so better use {@link #values()}
 * and use the method {@link XAttributeExternalImpl#getKey()} to get the key of
 * the attribute.
 * 
 * @author F. Mannhardt
 * 
 */
class MapDBAttributeMap implements XAttributeMap {

	private final ExternalAttributable owner;
	private final MapDBStore store;

	private NavigableMap<Long, XAttributeExternalImpl> subMap;

	/**
	 * Create a new {@link MapDBAttributeMap} for the
	 * {@link ExternalAttributable} in the {@link ExternalStore}. This class
	 * provides a view on the data stored, so no data is copied upon creating.
	 * 
	 * @param owner
	 * @param store
	 */
	public MapDBAttributeMap(ExternalAttributable owner, MapDBStore store) {
		this.owner = owner;
		this.store = store;		
	}
	
	final ConcurrentNavigableMap<Long, XAttributeExternalImpl> getInternalFullMap() {
		return store.getMapStorage();
	}
	
	final NavigableMap<Long, XAttributeExternalImpl> getInternalSubMap() {
		if (subMap == null) {
			subMap = getInternalFullMap().subMap(store.lowestCompositeKey(owner), true,
					store.highestCompositeKey(owner), true);
		}
		return subMap;
	}

	final MapDBStore getStore() {
		return store;
	}

	final ExternalAttributable getOwner() {
		return owner;
	}

	private final Integer safeGetKey(Object key) {
		if (key instanceof String) {
			return store.getInternalAttributeKey(owner, (String) key);
		} else {
			return null;
		}
	}

	@Override
	public void clear() {
		getInternalSubMap().clear();
	}

	@Override
	public final boolean containsKey(Object o) {
		Integer key = safeGetKey(o);
		if (key == null) {
			return false;
		} else {
			return getInternalFullMap().containsKey(MapDBStore.getCompositeKey(owner, key));
		}
	}

	@Override
	public final boolean containsValue(Object value) {
		return getInternalSubMap().containsValue(value);
	}

	@Override
	public final Set<Entry<String, XAttribute>> entrySet() {
		return new MapDBEntrySet(this);
	}

	@Override
	public XAttribute get(Object o) {
		Integer key = safeGetKey(o);
		if (key == null) {
			return null;
		} else {
			Long compositeKey = MapDBStore.getCompositeKey(owner, key);
			return XAttributeExternalImpl.decorate(getInternalFullMap().get(compositeKey),
					key, store);
		}
	}

	@Override
	public final boolean isEmpty() {
		return getInternalSubMap().isEmpty();
	}

	@Override
	public final Set<String> keySet() {
		return new MapDBKeySet(this);
	}

	@Override
	public XAttribute put(String key, XAttribute value) {
		if (key == null) {
			String attributeInfo = (value != null ? value.toString() : "NULL");
			throw new NullPointerException(
					"The MapDB implementation of XAttributeMap does not support the use of NULL as key. Are you sure that the following attribute is valid: "
							+ attributeInfo);
		}		
		XAttributeExternalImpl externalAttribute = XAttributeExternalImpl.convert(store, value);
		assert key.equals(externalAttribute.getKey()) : "Trying to put an XAttribute under different key than the one return by getKey";
		return getInternalFullMap().put(MapDBStore.getCompositeKey(owner, externalAttribute.getInternalKey()), externalAttribute);
	}

	@Override
	public final void putAll(Map<? extends String, ? extends XAttribute> m) {
		for (XAttribute attribute : m.values()) {
			put(attribute.getKey(), attribute);
		}
	}

	@Override
	public XAttribute remove(Object o) {
		Integer key = safeGetKey(o);
		if (key == null) {
			return null;
		} else {
			Long compositeKey = MapDBStore.getCompositeKey(owner, key);
			XAttributeExternalImpl removedAttribute = getInternalFullMap().remove(compositeKey);
			if (removedAttribute != null) {
				return XAttributeExternalImpl.decorate(removedAttribute, key, store);
			} else {
				return null;
			}
		}
	}

	@Override
	public final int size() {
		return getInternalSubMap().size();
	}

	@Override
	public final Collection<XAttribute> values() {
		return Collections2.transform(getInternalSubMap().entrySet(),
				new Function<Entry<Long, XAttributeExternalImpl>, XAttribute>() {

					@Override
					public XAttribute apply(Entry<Long, XAttributeExternalImpl> entry) {
						Long compositeKey = entry.getKey();
						XAttributeExternalImpl attribute = entry.getValue();
						int attributeKey = MapDBStore.getInternalAttributeKey(owner, compositeKey);						
						return XAttributeExternalImpl.decorate(attribute, attributeKey, store);
					}

				});
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#clone()
	 */
	@Override
	public Object clone() {
		// Return a real in-memory copy of this map
		XAttributeMap clone = new XAttributeMapImpl();
		for (XAttribute attr : this.values()) {
			clone.put(attr.getKey(), (XAttribute) attr.clone());
		}
		return clone;
	}

}