/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your option) any
 * later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */
package org.processmining.xeslite.external;

import java.io.Serializable;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicReferenceArray;
import java.util.concurrent.locks.LockSupport;

/**
 * @author F. Mannhardt
 * 
 */
final class StringPoolCASImpl implements Serializable, StringPool {

	private static final long serialVersionUID = -2582640172305237779L;

	private static final Integer INSERTING = new Integer(-1);

	private final int poolLimit;
	
	private final AtomicInteger keyCounter;
	private final ConcurrentMap<String, Integer> keyMap;
	private final AtomicReferenceArray<String> values;

	public StringPoolCASImpl(int poolLimit) {
		super();
		this.poolLimit = poolLimit;
		this.keyCounter = new AtomicInteger();
		this.keyMap = new ConcurrentHashMap<>();
		this.values = new AtomicReferenceArray<>(poolLimit);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.progressmining.xeslite.external.StringPool#put(java.lang.String)
	 */
	@Override
	public Integer put(String val) {

		if (keyCounter.get() == poolLimit) {
			throw new RuntimeException("Too many distinct keys in this XES file!");
		}

		// First check for existence, as we expect the same String to be used multiple times
		Integer index = keyMap.get(val);

		if (index != null) {
			while (index == INSERTING) {
				LockSupport.parkNanos(1L);
				index = keyMap.get(val);
			}
			return index;
		} else {
			// No yet present, insert a stub value as we don't want to waste indices
			index = keyMap.putIfAbsent(val, INSERTING);
			if (index == null) {
				// We reserved the spot and, therefore, the following operations are executed atomically
				// Take care that this block never fails, otherwise other threads trying the same put starve
				index = keyCounter.getAndIncrement();
				values.set(index, val);
				keyMap.put(val, index);
				return index;
			} else {
				// Another thread reserved the sport, wait until insert is visible to us
				while (index == INSERTING) {
					LockSupport.parkNanos(1L);
					index = keyMap.get(val);
				}
				return index;
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.progressmining.xeslite.external.StringPool#getIndex(java.lang.String)
	 */
	@Override
	public Integer getIndex(String val) {
		Integer index = keyMap.get(val);
		while (index != null && index == INSERTING) {
			LockSupport.parkNanos(1L);
			index = keyMap.get(val);
		}
		return index;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.progressmining.xeslite.external.StringPool#getValue(java.lang.Integer
	 * )
	 */
	@Override
	public String getValue(Integer index) {
		return values.get(index);
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return String.format("StringPoolCASImpl %s [keyCounter=%s, keyMap=%s, values=%s]", System.identityHashCode(this), keyCounter, keyMap, values);
	}

	public int getPoolLimit() {
		return poolLimit;
	}

}